from __future__ import annotations

from typing import Optional
from dataclasses import dataclass
from datetime import datetime, timedelta

from .awareness import ensure_aware


@dataclass(frozen=True)
class ContinuousTimeRange:
    start: datetime
    end: datetime

    def __post_init__(self):
        object.__setattr__(self, 'start', ensure_aware(self.start))
        object.__setattr__(self, 'end', ensure_aware(self.end))
        if self.start > self.end:
            raise ValueError("Start cannot be later than end.")

    def __contains__(self, moment: datetime) -> bool:
        return self.start <= moment <= self.end

    @property
    def length(self) -> timedelta:
        return self.end - self.start

    def __and__(self, other: ContinuousTimeRange) -> Optional[ContinuousTimeRange]:
        if type(self) != type(other):
            return NotImplemented
        if not self.intersects_with(other):
            return None
        return ContinuousTimeRange(max(self.start, other.start), min(self.end, other.end))

    def __add__(self, other) -> Optional[ContinuousTimeRange]:
        if type(self) != type(other):
            return NotImplemented
        if not self.intersects_with(other):
            return None
        return ContinuousTimeRange(min(self.start, other.start), max(self.end, other.end))

    def __ge__(self, other) -> bool:  # TODO TEST
        """
        Check if self (S) is a superset of other (O): S ⊇ O
        """
        return self.start <= other.start and other.end <= self.end

    def __le__(self, other) -> bool:  # TODO TEST
        """
        Check if self (S) is a subset of other (O): S ⊆ O
        """
        return other.start <= self.start and self.end <= other.end

    def intersects_with(self, other: ContinuousTimeRange) -> bool:
        return other.start in self or other.end in self or self <= other



from __future__ import annotations

from typing import overload, Optional, Set, FrozenSet
import operator
from functools import reduce
from datetime import datetime, date, timedelta

from .continuous import ContinuousTimeRange


class TimeRange:
    intervals: FrozenSet[ContinuousTimeRange] = frozenset()

    @overload
    def __init__(self):
        pass

    @overload
    def __init__(self, start: datetime, end: datetime):
        pass

    @overload
    def __init__(self, cls, start: datetime, duration: timedelta):
        pass

    def __init__(self, start: Optional[datetime] = None,
                 end: Optional[datetime] = None,
                 duration: Optional[timedelta] = None):
        if (start and not (end or duration)) or (end and duration):
            raise ValueError("Allowed combination of constructor parameters are:\n  "
                             "* Empty constructor\n  * `start` & `end`\n  * `start` & `duration`")
        if start:
            self.intervals = frozenset({ContinuousTimeRange(start, end or start + duration)})

    def __eq__(self, other: TimeRange) -> bool:
        return self.intervals == other.intervals

    def __repr__(self) -> str:
        return ' + '.join(
            [f"TimeRange(start={repr(p.start)}, end={repr(p.end)})" for p in self.intervals]
        ) or "TimeRange()"

    def __bool__(self) -> bool:
        """
        A TimeRange evaluates to True if its duration is grater than zero.
        """
        return len(self.intervals) != 0

    def __contains__(self, moment: datetime) -> bool:
        return any([moment in p for p in self.intervals])

    def __add__(self, other: TimeRange) -> TimeRange:
        """
        Find the union of the two TimeRange's: A ⋃ B.
        """
        if not isinstance(other, TimeRange):
            return NotImplemented
        intersectionless_periods: Set[ContinuousTimeRange] = set()
        for p in {*self.intervals, *other.intervals}:
            for per in intersectionless_periods:
                if p.intersects_with(per):
                    intersectionless_periods.discard(per)
                    intersectionless_periods.add(p + per)
                    break
            else:
                intersectionless_periods.add(p)
        result = TimeRange()
        result.intervals = intersectionless_periods
        return result

    def __and__(self, other: TimeRange) -> TimeRange:
        """
        Find the intersection of the two TimeRange's: A ⋂ B.
        """
        if not isinstance(other, TimeRange):
            return NotImplemented
        # Compute the Cartesian product and find all intersections
        intersections = {s & o for s in self.intervals for o in other.intervals}
        intersections.discard(None)
        # TODO on next line switch from [] compasion to {} comparison
        #  after TimeRange has been made hashable
        return reduce(operator.add, [TimeRange(t.start, t.end) for t in intersections], TimeRange())

    @property
    def length(self) -> timedelta:
        return sum([p.length for p in self.intervals], start=timedelta())

    @property
    def start(self) -> datetime:  # FIXME ca it be None if TimeRange is empty
        """
        Return the earliest moment in the TimeRange.
        """
        return min({p.start for p in self.intervals})

    @property
    def end(self) -> datetime:  # FIXME ca it be None if TimeRange is empty
        """
        Return the latest moment in the TimeRange.
        """
        return max({p.end for p in self.intervals})

    @property  # TODO TEST
    def start_date(self) -> date:  # FIXME can self.start be None?
        return self.start.date()

    @property  # TODO TEST
    def end_date(self) -> date:  # FIXME can self.end be None?
        return self.end.date()



from __future__ import annotations

from calendar import monthrange
from datetime import timedelta, date

from .timerange import TimeRange
from .date_range import daterange


class CalendarMonth(TimeRange):
    """
    Represent a calendar month.
    """

    def __init__(self, year: int, month: int):
        date_range = daterange(start=date(year, month, 1), end=self._last_date_of_month(year, month))
        super().__init__(start=date_range.start, end=date_range.end)
        self.year = year
        self.month = month

    def __repr__(self):
        return f'CalendarMonth(year={self.year}, month={self.month})'

    @property
    def next(self) -> CalendarMonth:
        """Return an instance of next month."""
        first_day_in_next_month = self.end.date() + timedelta(days=1)
        return CalendarMonth(first_day_in_next_month.year, first_day_in_next_month.month)

    @property
    def prev(self) -> CalendarMonth:
        """Return an instance of previous month."""
        last_day_in_previous_month = self.start.date() - timedelta(days=1)
        return CalendarMonth(last_day_in_previous_month.year, last_day_in_previous_month.month)

    @staticmethod
    def _last_date_of_month(year: int, month: int) -> date:
        _, last_day = monthrange(year, month)
        return date(year, month, last_day)



from typing import Type, Any


def ensure(condition: Any, exception_cls: Type[Exception], msg: str):
    if not condition:
        raise exception_cls(msg)


def preclude(condition: Any, exception_cls: Type[Exception], msg: str):
    ensure(not condition, exception_cls, msg)



from datetime import datetime, date, timedelta, time
from typing import Optional

from .timerange import TimeRange
from .utils import preclude


# TODO TEST
def daterange(start: date, end: Optional[date] = None, days: Optional[int] = None) -> TimeRange:
    preclude(end is not None and days is not None, TypeError, 'Only one of `end` `days` can be given.')
    if end is not None:
        end_time = end
    elif days is not None:
        end_time = end or start + timedelta(days=days)
    else:
        end_time = start
    return TimeRange(first_moment_in_day(start), last_moment_in_day(end_time))


def first_moment_in_day(d: date, /) -> datetime:
    return datetime.combine(d, time.min)


def last_moment_in_day(d: date, /) -> datetime:
    return datetime.combine(d, time.max)



from datetime import datetime

try:
    from django.utils.timezone import is_aware, make_aware
except ImportError:
    def is_aware(value: datetime) -> bool:
        return True


    def make_aware(value: datetime) -> datetime:
        return value


def ensure_aware(value: datetime) -> datetime:
    return value if is_aware(value) else make_aware(value)
